//--------------------------------------------------------------------
//
// bstree.cpp
//
// Linked implementation of the Binary Search Tree ADT
//
//--------------------------------------------------------------------

#include <new>
#include <iostream>
#include "bstree.h"

using namespace std;

//--------------------------------------------------------------------

template < class DT, class KF >
BSTreeNode<DT,KF>:: BSTreeNode ( const DT &nodeDataItem,
                                 BSTreeNode<DT,KF> *leftPtr,
                                 BSTreeNode<DT,KF> *rightPtr )

// Creates a binary search tree node containing data item elem, left
// child pointer leftPtr, and right child pointer rightPtr.

  : dataItem(nodeDataItem),
    left(leftPtr),
    right(rightPtr)

{}

//--------------------------------------------------------------------

template < class DT, class KF >
BSTree<DT,KF>:: BSTree ()

// Creates an empty tree.

  : root(nullptr)

{}

//--------------------------------------------------------------------

template < class DT, class KF >
BSTree<DT,KF>:: ~BSTree ()

// Frees the memory used by a tree.

{
    clear();
}

//--------------------------------------------------------------------

template < class DT, class KF >
void BSTree<DT,KF>:: insert ( const DT &newDataItem )
    throw ( bad_alloc )

// Inserts newDataItem into a tree. If an data item with the same key
// as newDataItem already exists in the tree, then updates that
// data item's data with newDataItem's data.

{
    insertSub(root,newDataItem);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
void BSTree<DT,KF>:: insertSub ( BSTreeNode<DT,KF> *&p,
                                 const DT &newDataItem   )
    throw ( bad_alloc )

// Recursive partner of the insert() function. Inserts newDataItem in
// the subtree pointed to by p.

{
    if ( p == nullptr ) // Insert
       p = new BSTreeNode<DT,KF>(newDataItem,nullptr,nullptr);
    else if ( newDataItem.key() < p->dataItem.key() )
       insertSub(p->left,newDataItem);                  // Search left
    else if ( newDataItem.key() > p->dataItem.key() )
       insertSub(p->right,newDataItem);                 // Search right
    else
       p->dataItem = newDataItem;                        // Update
}

//--------------------------------------------------------------------

template < class DT, class KF >
bool BSTree<DT,KF>:: retrieve ( KF searchKey, DT &searchDataItem ) const

// Searches a tree for the data item with key searchKey. If the data item
// is found, then copies the data item to searchDataItem and returns 1.
// Otherwise, returns 0 with searchDataItem undefined.

{
    return retrieveSub(root,searchKey,searchDataItem);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
bool BSTree<DT,KF>:: retrieveSub ( BSTreeNode<DT,KF> *p,
                                   KF searchKey,
                                   DT &searchDataItem     ) const

// Recursive partner of the retrieve() function. Searches the subtree
// pointed to by p.

{
    bool result;   // Result returned

    if ( p == nullptr )
       result = false;
    else if ( searchKey < p->dataItem.key() )
       result = retrieveSub(p->left,searchKey,searchDataItem);
    else if ( searchKey > p->dataItem.key() )
       result = retrieveSub(p->right,searchKey,searchDataItem);
    else
    {
       searchDataItem = p->dataItem;
       result = true;
    }

    return result;
}

//--------------------------------------------------------------------

template < class DT, class KF >
bool BSTree<DT,KF>:: remove ( KF deleteKey )

// Removes the data item with key deleteKey from a tree. If the
// data item is found, then deletes it from the tree and returns 1.
// Otherwise, returns 0.

{
    return removeSub(root,deleteKey);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
bool BSTree<DT,KF>:: removeSub ( BSTreeNode<DT,KF> *&p,
                                KF deleteKey           )

// Recursive partner of the remove() function. Searches the subtree
// pointed to by p for the node with key deleteKey. If this node is
// found, then does one of the following:
//
//    - If the node is missing one or more children, then links
//      around it and deletes it.
//
//    - Otherwise, uses cutRightmost to replace the data in the
//      node with the data in the rightmost descendant of the node's
//      left child and deletes that node.

{
    BSTreeNode<DT,KF> *delPtr;  // Pointer to node to delete
    int result;                 // Result returned

    if ( p == nullptr )
       result = false;                           // Search failed
    else if ( deleteKey < p->dataItem.key() )
       result = removeSub(p->left,deleteKey);    // Search left
    else if ( deleteKey > p->dataItem.key() )
       result = removeSub(p->right,deleteKey);   // Search right
    else
    {                                            // Found
       delPtr = p;
       if ( p->left == nullptr )
          p = p->right;                    // No left child
       else if ( p->right == nullptr )
          p = p->left;                     // No right child
       else
          cutRightmost(p->left,delPtr);    // Has both children
       delete delPtr;
       result = false;
    }

    return result;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
void BSTree<DT,KF>:: cutRightmost ( BSTreeNode<DT,KF> *&r,
                                    BSTreeNode<DT,KF> *&delPtr )

// Recursive partner of the removeSub() function. Traces down a
// sequence of right children. Copies the data from the last node in
// the sequence into the node pointed to delPtr, sets delPtr to point
// to the last node in the sequence, and links around this node.

{
    if ( r->right != nullptr )
       cutRightmost(r->right,delPtr);   // Continue down to the right
    else
    {
       delPtr->dataItem = r->dataItem;
       delPtr = r;
       r = r->left;
    }
}

//--------------------------------------------------------------------

template < class DT, class KF >
void BSTree<DT,KF>:: writeKeys () const

// Outputs the keys in a tree in ascending order.

{
    writeKeysSub(root);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
void BSTree<DT,KF>:: writeKeysSub ( BSTreeNode<DT,KF> *p ) const

// Recursive partner of the writeKeys() function. Outputs the keys in
// the subtree pointed to by p.

{
    if ( p != nullptr )
    {
       writeKeysSub(p->left);
       cout << p->dataItem.key() << endl;
       writeKeysSub(p->right);
    }
}

//--------------------------------------------------------------------

template < class DT, class KF >
void BSTree<DT,KF>:: clear ()

// Removes all the nodes from a tree.

{
    clearSub(root);
    root = nullptr;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
void BSTree<DT,KF>:: clearSub ( BSTreeNode<DT,KF> *p )

// Recursive partner of the clear() function. Clears the subtree
// pointed to by p.

{
    if ( p != nullptr)
    {
       clearSub(p->left);
       clearSub(p->right);
       delete p;
    }
}

//--------------------------------------------------------------------

template < class DT, class KF >
bool BSTree<DT,KF>:: isEmpty () const

// Returns 1 if a tree is empty. Otherwise returns 0.

{
    return ( root == nullptr );
}

//--------------------------------------------------------------------

template < class DT, class KF >
bool BSTree<DT,KF>:: isFull () const

// Returns true if a list is full. Otherwise, returns false.

{
    // This is a somewhat hacked way to test if the list is full.
    // If a node can be successfully allocated than the list is not
    // full.  If the allocation fails it is implied that there is no
    // more free memory therefore the list is full.
    
    DT testDataItem;
    BSTreeNode<DT,KF> *p;
    
    try
    {
        p = new BSTreeNode<DT,KF>(testDataItem, nullptr, nullptr);
    }
    catch ( bad_alloc &e )
    {
        return true;
    }

    delete p;
    return false;
}

//--------------------------------------------------------------------

template < class DT, class KF >
void BSTree<DT,KF>:: showStructure () const

// Outputs the keys in a binary search tree. The tree is output
// rotated counterclockwise 90 degrees from its conventional
// orientation using a "reverse" inorder traversal. This operation is
// intended for testing and debugging purposes only.

{
    if ( root == nullptr )
       cout << "Empty tree" << endl;
    else
    {
       cout << endl;
       showSub(root,1);
       cout << endl;
    }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
void BSTree<DT,KF>:: showSub ( BSTreeNode<DT,KF> *p,
                               int level             ) const

// Recursive partner of the showStructure() function. Outputs the
// subtree whose root node is pointed to by p. Parameter level is the
// level of this node within the tree.

{
     int j;   // Loop counter

     if ( p != nullptr )
     {
        showSub(p->right,level+1);         // Output right subtree
        for ( j = 0 ; j < level ; j++ )    // Tab over to level
            cout << "\t";
        cout << " " << p->dataItem.key();   // Output key
        if ( ( p->left != nullptr ) &&           // Output "connector"
             ( p->right != nullptr ) )
           cout << "<";
        else if ( p->right != nullptr )
           cout << "/";
        else if ( p->left != nullptr )
           cout << "\\";
        cout << endl;
        showSub(p->left,level+1);          // Output left subtree
    }
}

//--------------------------------------------------------------------
//
//                         Helper operations
//
//--------------------------------------------------------------------

template < class DT, class KF >
void BSTree<DT,KF>:: writeLessThan ( KF searchKey ) const

// Outputs the keys in a tree that are less than searchKey.

{
   writeLTSub(root,searchKey);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
void BSTree<DT,KF>:: writeLTSub ( BSTreeNode<DT,KF> *p,
                                  KF searchKey          ) const

// Recursive partner of the writeLessThan() function. Outputs the keys
// in the subtree pointed to by p that are less than searchKey.

{
    if ( p != nullptr )
    {
       writeLTSub(p->left,searchKey);
       if ( searchKey > p->dataItem.key() )
       {
          cout << p->dataItem.key() << endl;
          writeLTSub(p->right,searchKey);
       }
    }
}


//--------------------------------------------------------------------

template < class DT, class KF >
int BSTree<DT,KF>:: getHeight () const

// Returns the height of a tree.

{
    return getHeightSub(root);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

template < class DT, class KF >
int BSTree<DT,KF>:: getHeightSub ( BSTreeNode<DT,KF> *p ) const

// Recursive partner of the height() function. Returns the height of
// the subtree pointed to by p -- that is, the length of the longest
// path from the node pointed to by p to any leaf node.

{
    int l,        // Length of the longest path thru the left child
        r,        // Length of the longest path thru the right child
        result;   // Result returned

     if ( p == nullptr )
        result = 0;                    // Leaf node
     else
     {
        l = getHeightSub(p->left) + 1;    // Get height of left subtree
        r = getHeightSub(p->right) + 1;   // Get height of right subtree
        if ( l >= r )                  // Return larger
           result = l;
        else
           result = r;
     }

     return result;
}
